# -*- coding: utf-8 -*-
"""
Utility functions used systemwide.

"""

from __future__ import absolute_import

import binascii
import collections
import errno
import grp
import logging
import os
import random
import re
import shutil
import smtplib
import stat
import string
import sys
import time
import tempfile
import threading
from six.moves.urllib import parse as urlparse

from galaxy.util import json
from datetime import datetime

from six import PY3
from six import string_types, text_type
from six.moves import xrange
from six.moves import email_mime_text
from six.moves import zip

from os.path import relpath
from hashlib import md5

import docutils.core
import docutils.writers.html4css1

from xml.etree import ElementTree, ElementInclude

from .inflection import Inflector, English
inflector = Inflector(English)

if PY3:
    def list_map(f, input):
        return list(map(f, input))
else:
    list_map = map


log = logging.getLogger(__name__)
_lock = threading.RLock()

CHUNK_SIZE = 65536  # 64k

DATABASE_MAX_STRING_SIZE = 32768
DATABASE_MAX_STRING_SIZE_PRETTY = '32K'

gzip_magic = '\037\213'
bz2_magic = 'BZh'
DEFAULT_ENCODING = os.environ.get('GALAXY_DEFAULT_ENCODING', 'utf-8')
NULL_CHAR = '\000'
BINARY_CHARS = [ NULL_CHAR ]


def remove_protocol_from_url( url ):
    """ Supplied URL may be null, if not ensure http:// or https://
    etc... is stripped off.
    """
    if url is None:
        return url

    # We have a URL
    if url.find( '://' ) > 0:
        new_url = url.split( '://' )[1]
    else:
        new_url = url
    return new_url.rstrip( '/' )


def is_binary( value, binary_chars=None ):
    """
    File is binary if it contains a null-byte by default (e.g. behavior of grep, etc.).
    This may fail for utf-16 files, but so would ASCII encoding.
    >>> is_binary( string.printable )
    False
    >>> is_binary( '\\xce\\x94' )
    False
    >>> is_binary( '\\000' )
    True
    """
    if binary_chars is None:
        binary_chars = BINARY_CHARS
    for binary_char in binary_chars:
        if binary_char in value:
            return True
    return False


def is_uuid( value ):
    """
    This method returns True if value is a UUID, otherwise False.
    >>> is_uuid( "123e4567-e89b-12d3-a456-426655440000" )
    True
    >>> is_uuid( "0x3242340298902834" )
    False
    """
    uuid_re = re.compile( "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}" )
    if re.match( uuid_re, str( value ) ):
        return True
    else:
        return False


def get_charset_from_http_headers( headers, default=None ):
    rval = headers.get('content-type', None )
    if rval and 'charset=' in rval:
        rval = rval.split('charset=')[-1].split(';')[0].strip()
        if rval:
            return rval
    return default


def synchronized(func):
    """This wrapper will serialize access to 'func' to a single thread. Use it as a decorator."""
    def caller(*params, **kparams):
        _lock.acquire(True)  # Wait
        try:
            return func(*params, **kparams)
        finally:
            _lock.release()
    return caller


def file_iter(fname, sep=None):
    """
    This generator iterates over a file and yields its lines
    splitted via the C{sep} parameter. Skips empty lines and lines starting with
    the C{#} character.

    >>> lines = [ line for line in file_iter(__file__) ]
    >>> len(lines) !=  0
    True
    """
    for line in open(fname):
        if line and line[0] != '#':
            yield line.split(sep)


def file_reader( fp, chunk_size=CHUNK_SIZE ):
    """This generator yields the open fileobject in chunks (default 64k). Closes the file at the end"""
    while 1:
        data = fp.read(chunk_size)
        if not data:
            break
        yield data
    fp.close()


def unique_id(KEY_SIZE=128):
    """
    Generates an unique id

    >>> ids = [ unique_id() for i in range(1000) ]
    >>> len(set(ids))
    1000
    """
    random_bits = text_type(random.getrandbits(KEY_SIZE)).encode("UTF-8")
    return md5(random_bits).hexdigest()


def parse_xml( fname ):
    """Returns a parsed xml tree"""
    # handle deprecation warning for XMLParsing a file with DOCTYPE
    class DoctypeSafeCallbackTarget( ElementTree.TreeBuilder ):
        def doctype( *args ):
            pass
    tree = ElementTree.ElementTree()
    root = tree.parse( fname, parser=ElementTree.XMLParser( target=DoctypeSafeCallbackTarget() ) )
    ElementInclude.include( root )
    return tree


def parse_xml_string(xml_string):
    tree = ElementTree.fromstring(xml_string)
    return tree


def xml_to_string( elem, pretty=False ):
    """Returns a string from an xml tree"""
    if pretty:
        elem = pretty_print_xml( elem )
    try:
        return ElementTree.tostring( elem )
    except TypeError as e:
        # we assume this is a comment
        if hasattr( elem, 'text' ):
            return "<!-- %s -->\n" % ( elem.text )
        else:
            raise e


def xml_element_compare( elem1, elem2 ):
    if not isinstance( elem1, dict ):
        elem1 = xml_element_to_dict( elem1 )
    if not isinstance( elem2, dict ):
        elem2 = xml_element_to_dict( elem2 )
    return elem1 == elem2


def xml_element_list_compare( elem_list1, elem_list2 ):
    return [ xml_element_to_dict( elem ) for elem in elem_list1  ] == [ xml_element_to_dict( elem ) for elem in elem_list2  ]


def xml_element_to_dict( elem ):
    rval = {}
    if elem.attrib:
        rval[ elem.tag ] = {}
    else:
        rval[ elem.tag ] = None

    sub_elems = list( elem )
    if sub_elems:
        sub_elem_dict = dict()
        for sub_sub_elem_dict in map( xml_element_to_dict, sub_elems ):
            for key, value in sub_sub_elem_dict.iteritems():
                if key not in sub_elem_dict:
                    sub_elem_dict[ key ] = []
                sub_elem_dict[ key ].append( value )
        for key, value in sub_elem_dict.iteritems():
            if len( value ) == 1:
                rval[ elem.tag ][ key ] = value[0]
            else:
                rval[ elem.tag ][ key ] = value
    if elem.attrib:
        for key, value in elem.attrib.iteritems():
            rval[ elem.tag ][ "@%s" % key ] = value

    if elem.text:
        text = elem.text.strip()
        if text and sub_elems or elem.attrib:
            rval[ elem.tag ][ '#text' ] = text
        else:
            rval[ elem.tag ] = text

    return rval


def pretty_print_xml( elem, level=0 ):
    pad = '    '
    i = "\n" + level * pad
    if len( elem ):
        if not elem.text or not elem.text.strip():
            elem.text = i + pad + pad
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for e in elem:
            pretty_print_xml( e, level + 1 )
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and ( not elem.tail or not elem.tail.strip() ):
            elem.tail = i + pad
    return elem


def get_file_size( value, default=None ):
    try:
        # try built-in
        return os.path.getsize( value )
    except:
        try:
            # try built-in one name attribute
            return os.path.getsize( value.name )
        except:
            try:
                # try tell() of end of object
                offset = value.tell()
                value.seek( 0, 2 )
                rval = value.tell()
                value.seek( offset )
                return rval
            except:
                # return default value
                return default


def shrink_stream_by_size( value, size, join_by="..", left_larger=True, beginning_on_size_error=False, end_on_size_error=False ):
    rval = ''
    if get_file_size( value ) > size:
        start = value.tell()
        len_join_by = len( join_by )
        min_size = len_join_by + 2
        if size < min_size:
            if beginning_on_size_error:
                rval = value.read( size )
                value.seek( start )
                return rval
            elif end_on_size_error:
                value.seek( -size, 2 )
                rval = value.read( size )
                value.seek( start )
                return rval
            raise ValueError( 'With the provided join_by value (%s), the minimum size value is %i.' % ( join_by, min_size ) )
        left_index = right_index = int( ( size - len_join_by ) / 2 )
        if left_index + right_index + len_join_by < size:
            if left_larger:
                left_index += 1
            else:
                right_index += 1
        rval = value.read( left_index ) + join_by
        value.seek( -right_index, 2 )
        rval += value.read( right_index )
    else:
        while True:
            data = value.read( CHUNK_SIZE )
            if not data:
                break
            rval += data
    return rval


def shrink_string_by_size( value, size, join_by="..", left_larger=True, beginning_on_size_error=False, end_on_size_error=False ):
    if len( value ) > size:
        len_join_by = len( join_by )
        min_size = len_join_by + 2
        if size < min_size:
            if beginning_on_size_error:
                return value[:size]
            elif end_on_size_error:
                return value[-size:]
            raise ValueError( 'With the provided join_by value (%s), the minimum size value is %i.' % ( join_by, min_size ) )
        left_index = right_index = int( ( size - len_join_by ) / 2 )
        if left_index + right_index + len_join_by < size:
            if left_larger:
                left_index += 1
            else:
                right_index += 1
        value = "%s%s%s" % ( value[:left_index], join_by, value[-right_index:] )
    return value


def pretty_print_time_interval( time=False, precise=False ):
    """
    Get a datetime object or a int() Epoch timestamp and return a
    pretty string like 'an hour ago', 'Yesterday', '3 months ago',
    'just now', etc
    credit: http://stackoverflow.com/questions/1551382/user-friendly-time-format-in-python
    """
    now = datetime.now()
    if type( time ) is int:
        diff = now - datetime.fromtimestamp( time )
    elif isinstance( time, datetime ):
        diff = now - time
    elif isinstance( time, string_types ):
        time = datetime.strptime( time, "%Y-%m-%dT%H:%M:%S.%f" )
        diff = now - time
    else:
        diff = now - now
    second_diff = diff.seconds
    day_diff = diff.days

    if day_diff < 0:
        return ''

    if precise:
        if day_diff == 0:
                if second_diff < 10:
                    return "just now"
                if second_diff < 60:
                    return str(second_diff) + " seconds ago"
                if second_diff < 120:
                    return "a minute ago"
                if second_diff < 3600:
                    return str(second_diff / 60) + " minutes ago"
                if second_diff < 7200:
                    return "an hour ago"
                if second_diff < 86400:
                    return str(second_diff / 3600) + " hours ago"
        if day_diff == 1:
            return "yesterday"
        if day_diff < 7:
            return str( day_diff ) + " days ago"
        if day_diff < 31:
            return str( day_diff / 7 ) + " weeks ago"
        if day_diff < 365:
            return str( day_diff / 30 ) + " months ago"
        return str( day_diff / 365 ) + " years ago"
    else:
        if day_diff == 0:
            return "today"
        if day_diff == 1:
            return "yesterday"
        if day_diff < 7:
            return "less than a week"
        if day_diff < 31:
            return "less than a month"
        if day_diff < 365:
            return "less than a year"
        return "a few years ago"


def pretty_print_json(json_data, is_json_string=False):
    if is_json_string:
        json_data = json.loads(json_data)
    return json.dumps(json_data, sort_keys=True, indent=4)

# characters that are valid
valid_chars = set(string.ascii_letters + string.digits + " -=_.()/+*^,:?!")

# characters that are allowed but need to be escaped
mapped_chars = { '>': '__gt__',
                 '<': '__lt__',
                 "'": '__sq__',
                 '"': '__dq__',
                 '[': '__ob__',
                 ']': '__cb__',
                 '{': '__oc__',
                 '}': '__cc__',
                 '@': '__at__',
                 '\n': '__cn__',
                 '\r': '__cr__',
                 '\t': '__tc__',
                 '#': '__pd__'}


def restore_text( text, character_map=mapped_chars ):
    """Restores sanitized text"""
    if not text:
        return text
    for key, value in character_map.items():
        text = text.replace(value, key)
    return text


def sanitize_text( text, valid_characters=valid_chars, character_map=mapped_chars, invalid_character='X' ):
    """
    Restricts the characters that are allowed in text; accepts both strings
    and lists of strings; non-string entities will be cast to strings.
    """
    if isinstance( text, list ):
        return list_map( lambda x: sanitize_text( x, valid_characters=valid_characters, character_map=character_map, invalid_character=invalid_character ), text )
    if not isinstance( text, string_types ):
        text = smart_str( text )
    return _sanitize_text_helper( text, valid_characters=valid_characters, character_map=character_map )


def _sanitize_text_helper( text, valid_characters=valid_chars, character_map=mapped_chars, invalid_character='X' ):
    """Restricts the characters that are allowed in a string"""

    out = []
    for c in text:
        if c in valid_characters:
            out.append(c)
        elif c in character_map:
            out.append( character_map[c] )
        else:
            out.append( invalid_character )  # makes debugging easier
    return ''.join(out)


def sanitize_lists_to_string( values, valid_characters=valid_chars, character_map=mapped_chars, invalid_character='X'  ):
    if isinstance( values, list ):
        rval = []
        for value in values:
            rval.append( sanitize_lists_to_string( value,
                                                   valid_characters=valid_characters,
                                                   character_map=character_map,
                                                   invalid_character=invalid_character ) )
        values = ",".join( rval )
    else:
        values = sanitize_text( values, valid_characters=valid_characters, character_map=character_map, invalid_character=invalid_character )
    return values


def sanitize_param( value, valid_characters=valid_chars, character_map=mapped_chars, invalid_character='X' ):
    """Clean incoming parameters (strings or lists)"""
    if isinstance( value, string_types ):
        return sanitize_text( value, valid_characters=valid_characters, character_map=character_map, invalid_character=invalid_character )
    elif isinstance( value, list ):
        return list_map( lambda x: sanitize_text( x, valid_characters=valid_characters, character_map=character_map, invalid_character=invalid_character ), value )
    else:
        raise Exception('Unknown parameter type (%s)' % ( type( value ) ))

valid_filename_chars = set( string.ascii_letters + string.digits + '_.' )
invalid_filenames = [ '', '.', '..' ]


def sanitize_for_filename( text, default=None ):
    """
    Restricts the characters that are allowed in a filename portion; Returns default value or a unique id string if result is not a valid name.
    Method is overly aggressive to minimize possible complications, but a maximum length is not considered.
    """
    out = []
    for c in text:
        if c in valid_filename_chars:
            out.append( c )
        else:
            out.append( '_' )
    out = ''.join( out )
    if out in invalid_filenames:
        if default is None:
            return sanitize_for_filename( str( unique_id() ) )
        return default
    return out


def mask_password_from_url( url ):
    """
    Masks out passwords from connection urls like the database connection in galaxy.ini

    >>> mask_password_from_url( 'sqlite+postgresql://user:password@localhost/' )
    'sqlite+postgresql://user:********@localhost/'
    >>> mask_password_from_url( 'amqp://user:amqp@localhost' )
    'amqp://user:********@localhost'
    >>> mask_password_from_url( 'amqp://localhost')
    'amqp://localhost'
    """
    split = urlparse.urlsplit(url)
    if split.password:
        if url.count(split.password) == 1:
            url = url.replace(split.password, "********")
        else:
            # This can manipulate the input other than just masking password,
            # so the previous string replace method is preferred when the
            # password doesn't appear twice in the url
            split = split._replace(netloc=split.netloc.replace("%s:%s" % (split.username, split.password), '%s:********' % split.username))
            url = urlparse.urlunsplit(split)
    return url


def ready_name_for_url( raw_name ):
    """ General method to convert a string (i.e. object name) to a URL-ready
    slug.

    >>> ready_name_for_url( "My Cool Object" )
    'My-Cool-Object'
    >>> ready_name_for_url( "!My Cool Object!" )
    'My-Cool-Object'
    >>> ready_name_for_url( "Hello₩◎ґʟⅾ" )
    'Hello'
    """

    # Replace whitespace with '-'
    slug_base = re.sub( "\s+", "-", raw_name )
    # Remove all non-alphanumeric characters.
    slug_base = re.sub( "[^a-zA-Z0-9\-]", "", slug_base )
    # Remove trailing '-'.
    if slug_base.endswith('-'):
        slug_base = slug_base[:-1]
    return slug_base


def in_directory( file, directory, local_path_module=os.path ):
    """
    Return true, if the common prefix of both is equal to directory
    e.g. /a/b/c/d.rst and directory is /a/b, the common prefix is /a/b
    """

    # Make both absolute.
    directory = local_path_module.abspath(directory)
    file = local_path_module.abspath(file)
    return local_path_module.commonprefix([file, directory]) == directory


def merge_sorted_iterables( operator, *iterables ):
    """

    >>> operator = lambda x: x
    >>> list( merge_sorted_iterables( operator, [1,2,3], [4,5] ) )
    [1, 2, 3, 4, 5]
    >>> list( merge_sorted_iterables( operator, [4, 5], [1,2,3] ) )
    [1, 2, 3, 4, 5]
    >>> list( merge_sorted_iterables( operator, [1, 4, 5], [2], [3] ) )
    [1, 2, 3, 4, 5]
    """
    first_iterable = iterables[ 0 ]
    if len( iterables ) == 1:
        for el in first_iterable:
            yield el
    else:
        for el in __merge_two_sorted_iterables(
            operator,
            iter( first_iterable ),
            merge_sorted_iterables( operator, *iterables[ 1: ] )
        ):
            yield el


def __merge_two_sorted_iterables( operator, iterable1, iterable2 ):
    unset = object()
    continue_merge = True
    next_1 = unset
    next_2 = unset
    while continue_merge:
        try:
            if next_1 is unset:
                next_1 = next( iterable1 )
            if next_2 is unset:
                next_2 = next( iterable2 )
            if operator( next_2 ) < operator( next_1 ):
                yield next_2
                next_2 = unset
            else:
                yield next_1
                next_1 = unset
        except StopIteration:
            continue_merge = False
    if next_1 is not unset:
        yield next_1
    if next_2 is not unset:
        yield next_2
    for el in iterable1:
        yield el
    for el in iterable2:
        yield el


class Params( object ):
    """
    Stores and 'sanitizes' parameters. Alphanumeric characters and the
    non-alphanumeric ones that are deemed safe are let to pass through (see L{valid_chars}).
    Some non-safe characters are escaped to safe forms for example C{>} becomes C{__lt__}
    (see L{mapped_chars}). All other characters are replaced with C{X}.

    Operates on string or list values only (HTTP parameters).

    >>> values = { 'status':'on', 'symbols':[  'alpha', '<>', '$rm&#!' ]  }
    >>> par = Params(values)
    >>> par.status
    'on'
    >>> par.value == None      # missing attributes return None
    True
    >>> par.get('price', 0)
    0
    >>> par.symbols            # replaces unknown symbols with X
    ['alpha', '__lt____gt__', 'XrmX__pd__!']
    >>> sorted(par.flatten())  # flattening to a list
    [('status', 'on'), ('symbols', 'XrmX__pd__!'), ('symbols', '__lt____gt__'), ('symbols', 'alpha')]
    """

    # is NEVER_SANITIZE required now that sanitizing for tool parameters can be controlled on a per parameter basis and occurs via InputValueWrappers?
    NEVER_SANITIZE = ['file_data', 'url_paste', 'URL', 'filesystem_paths']

    def __init__( self, params, sanitize=True ):
        if sanitize:
            for key, value in params.items():
                # sanitize check both ungrouped and grouped parameters by
                # name. Anything relying on NEVER_SANITIZE should be
                # changed to not require this and NEVER_SANITIZE should be
                # removed.
                if key not in self.NEVER_SANITIZE and True not in [ key.endswith( "|%s" % nonsanitize_parameter ) for
                                                                    nonsanitize_parameter in self.NEVER_SANITIZE ]:
                    self.__dict__[ key ] = sanitize_param( value )
                else:
                    self.__dict__[ key ] = value
        else:
            self.__dict__.update(params)

    def flatten(self):
        """
        Creates a tuple list from a dict with a tuple/value pair for every value that is a list
        """
        flat = []
        for key, value in self.__dict__.items():
            if isinstance(value, list):
                for v in value:
                    flat.append( (key, v) )
            else:
                flat.append( (key, value) )
        return flat

    def __getattr__(self, name):
        """This is here to ensure that we get None for non existing parameters"""
        return None

    def get(self, key, default):
        return self.__dict__.get(key, default)

    def __str__(self):
        return '%s' % self.__dict__

    def __len__(self):
        return len(self.__dict__)

    def __iter__(self):
        return iter(self.__dict__)

    def update(self, values):
        self.__dict__.update(values)


def rst_to_html( s ):
    """Convert a blob of reStructuredText to HTML"""
    log = logging.getLogger( "docutils" )

    class FakeStream( object ):
        def write( self, str ):
            if len( str ) > 0 and not str.isspace():
                log.warn( str )

    settings_overrides = {
        "embed_stylesheet": False,
        "template": os.path.join(os.path.dirname(__file__), "docutils_template.txt"),
        "warning_stream": FakeStream(),
        "doctitle_xform": False,  # without option, very different rendering depending on
                                  # number of sections in help content.
    }

    return unicodify( docutils.core.publish_string( s,
                      writer=docutils.writers.html4css1.Writer(),
                      settings_overrides=settings_overrides ) )


def xml_text(root, name=None):
    """Returns the text inside an element"""
    if name is not None:
        # Try attribute first
        val = root.get(name)
        if val:
            return val
        # Then try as element
        elem = root.find(name)
    else:
        elem = root
    if elem is not None and elem.text:
        text = ''.join(elem.text.splitlines())
        return text.strip()
    # No luck, return empty string
    return ''

# asbool implementation pulled from PasteDeploy
truthy = frozenset(['true', 'yes', 'on', 'y', 't', '1'])
falsy = frozenset(['false', 'no', 'off', 'n', 'f', '0'])


def asbool(obj):
    if isinstance(obj, string_types):
        obj = obj.strip().lower()
        if obj in truthy:
            return True
        elif obj in falsy:
            return False
        else:
            raise ValueError("String is not true/false: %r" % obj)
    return bool(obj)


def string_as_bool( string ):
    if str( string ).lower() in ( 'true', 'yes', 'on' ):
        return True
    else:
        return False


def string_as_bool_or_none( string ):
    """
    Returns True, None or False based on the argument:
        True if passed True, 'True', 'Yes', or 'On'
        None if passed None or 'None'
        False otherwise

    Note: string comparison is case-insensitive so lowecase versions of those
    function equivalently.
    """
    string = str( string ).lower()
    if string in ( 'true', 'yes', 'on' ):
        return True
    elif string == 'none':
        return None
    else:
        return False


def listify( item, do_strip=False ):
    """
    Make a single item a single item list, or return a list if passed a
    list.  Passing a None returns an empty list.
    """
    if not item:
        return []
    elif isinstance( item, list ):
        return item
    elif isinstance( item, string_types ) and item.count( ',' ):
        if do_strip:
            return [token.strip() for token in item.split( ',' )]
        else:
            return item.split( ',' )
    else:
        return [ item ]


def commaify(amount):
    orig = amount
    new = re.sub("^(-?\d+)(\d{3})", '\g<1>,\g<2>', amount)
    if orig == new:
        return new
    else:
        return commaify(new)


def roundify(amount, sfs=2):
    """
    Take a number in string form and truncate to 'sfs' significant figures.
    """
    if len(amount) <= sfs:
        return amount
    else:
        return amount[0:sfs] + '0' * (len(amount) - sfs)


def unicodify( value, encoding=DEFAULT_ENCODING, error='replace', default=None ):
    """
    Returns a unicode string or None
    """

    if isinstance( value, text_type ):
        return value
    try:
        return text_type( str( value ), encoding, error )
    except:
        return default


def smart_str(s, encoding='utf-8', strings_only=False, errors='strict'):
    """
    Returns a bytestring version of 's', encoded as specified in 'encoding'.

    If strings_only is True, don't convert (some) non-string-like objects.

    Adapted from an older, simpler version of django.utils.encoding.smart_str.
    """
    if strings_only and isinstance(s, (type(None), int)):
        return s
    if not isinstance(s, string_types):
        try:
            return str(s)
        except UnicodeEncodeError:
            return text_type(s).encode(encoding, errors)
    elif isinstance(s, text_type):
        return s.encode(encoding, errors)
    elif s and encoding != 'utf-8':
        return s.decode('utf-8', errors).encode(encoding, errors)
    else:
        return s


def object_to_string( obj ):
    return binascii.hexlify( obj )


def string_to_object( s ):
    return binascii.unhexlify( s )


class ParamsWithSpecs( collections.defaultdict ):
    """
    """

    def __init__( self, specs=None, params=None ):
        self.specs = specs or dict()
        self.params = params or dict()
        for name, value in self.params.items():
            if name not in self.specs:
                self._param_unknown_error( name )
            if 'map' in self.specs[ name ]:
                try:
                    self.params[ name ] = self.specs[ name ][ 'map' ]( value )
                except Exception:
                    self._param_map_error( name, value )
            if 'valid' in self.specs[ name ]:
                if not self.specs[ name ][ 'valid' ]( value ):
                    self._param_vaildation_error( name, value )

        self.update( self.params )

    def __missing__( self, name ):
        return self.specs[ name ][ 'default' ]

    def __getattr__( self, name ):
        return self[ name ]

    def _param_unknown_error( self, name ):
        raise NotImplementedError()

    def _param_map_error( self, name, value ):
        raise NotImplementedError()

    def _param_vaildation_error( self, name, value ):
        raise NotImplementedError()


def compare_urls( url1, url2, compare_scheme=True, compare_hostname=True, compare_path=True ):
    url1 = urlparse.urlparse( url1 )
    url2 = urlparse.urlparse( url2 )
    if compare_scheme and url1.scheme and url2.scheme and url1.scheme != url2.scheme:
        return False
    if compare_hostname and url1.hostname and url2.hostname and url1.hostname != url2.hostname:
        return False
    if compare_path and url1.path and url2.path and url1.path != url2.path:
        return False
    return True


def read_dbnames(filename):
    """ Read build names from file """
    class DBNames( list ):
        default_value = "?"
        default_name = "unspecified (?)"
    db_names = DBNames()
    try:
        ucsc_builds = {}
        man_builds = []  # assume these are integers
        name_to_db_base = {}
        if filename is None:
            # Should only be happening with the galaxy.tools.parameters.basic:GenomeBuildParameter docstring unit test
            filename = os.path.join( 'tool-data', 'shared', 'ucsc', 'builds.txt.sample' )
        for line in open(filename):
            try:
                if line[0:1] == "#":
                    continue
                fields = line.replace("\r", "").replace("\n", "").split("\t")
                # Special case of unspecified build is at top of list
                if fields[0] == "?":
                    db_names.insert(0, (fields[0], fields[1]))
                    continue
                try:  # manual build (i.e. microbes)
                    int(fields[0])
                    man_builds.append((fields[1], fields[0]))
                except:  # UCSC build
                    db_base = fields[0].rstrip('0123456789')
                    if db_base not in ucsc_builds:
                        ucsc_builds[db_base] = []
                        name_to_db_base[fields[1]] = db_base
                    # we want to sort within a species numerically by revision number
                    build_rev = re.compile(r'\d+$')
                    try:
                        build_rev = int(build_rev.findall(fields[0])[0])
                    except:
                        build_rev = 0
                    ucsc_builds[db_base].append((build_rev, fields[0], fields[1]))
            except:
                continue
        sort_names = name_to_db_base.keys()
        sort_names.sort()
        for name in sort_names:
            db_base = name_to_db_base[name]
            ucsc_builds[db_base].sort()
            ucsc_builds[db_base].reverse()
            ucsc_builds[db_base] = [(build, name) for _, build, name in ucsc_builds[db_base]]
            db_names = DBNames( db_names + ucsc_builds[db_base] )
        if len( db_names ) > 1 and len( man_builds ) > 0:
            db_names.append( ( db_names.default_value, '----- Additional Species Are Below -----' ) )
        man_builds.sort()
        man_builds = [(build, name) for name, build in man_builds]
        db_names = DBNames( db_names + man_builds )
    except Exception as e:
        log.error( "ERROR: Unable to read builds file: %s", e )
    if len(db_names) < 1:
        db_names = DBNames( [( db_names.default_value, db_names.default_name )] )
    return db_names


def read_build_sites( filename, check_builds=True ):
    """ read db names to ucsc mappings from file, this file should probably be merged with the one above """
    build_sites = []
    try:
        for line in open(filename):
            try:
                if line[0:1] == "#":
                    continue
                fields = line.replace("\r", "").replace("\n", "").split("\t")
                site_name = fields[0]
                site = fields[1]
                if check_builds:
                    site_builds = fields[2].split(",")
                    site_dict = {'name': site_name, 'url': site, 'builds': site_builds}
                else:
                    site_dict = {'name': site_name, 'url': site}
                build_sites.append( site_dict )
            except:
                continue
    except:
        log.error( "ERROR: Unable to read builds for site file %s", filename )
    return build_sites


def relativize_symlinks( path, start=None, followlinks=False):
    for root, dirs, files in os.walk( path, followlinks=followlinks ):
        rel_start = None
        for file_name in files:
            symlink_file_name = os.path.join( root, file_name )
            if os.path.islink( symlink_file_name ):
                symlink_target = os.readlink( symlink_file_name )
                if rel_start is None:
                    if start is None:
                        rel_start = root
                    else:
                        rel_start = start
                rel_path = relpath( symlink_target, rel_start )
                os.remove( symlink_file_name )
                os.symlink( rel_path, symlink_file_name )


def stringify_dictionary_keys( in_dict ):
    # returns a new dictionary
    # changes unicode keys into strings, only works on top level (does not recurse)
    # unicode keys are not valid for expansion into keyword arguments on method calls
    out_dict = {}
    for key, value in in_dict.iteritems():
        out_dict[ str( key ) ] = value
    return out_dict


def recursively_stringify_dictionary_keys( d ):
    if isinstance(d, dict):
        return dict([(k.encode( DEFAULT_ENCODING ), recursively_stringify_dictionary_keys(v)) for k, v in d.iteritems()])
    elif isinstance(d, list):
        return [recursively_stringify_dictionary_keys(x) for x in d]
    else:
        return d


def mkstemp_ln( src, prefix='mkstemp_ln_' ):
    """
    From tempfile._mkstemp_inner, generate a hard link in the same dir with a
    random name.  Created so we can persist the underlying file of a
    NamedTemporaryFile upon its closure.
    """
    dir = os.path.dirname(src)
    names = tempfile._get_candidate_names()
    for seq in xrange(tempfile.TMP_MAX):
        name = names.next()
        file = os.path.join(dir, prefix + name)
        try:
            os.link( src, file )
            return (os.path.abspath(file))
        except OSError as e:
            if e.errno == errno.EEXIST:
                continue  # try again
            raise
    raise IOError(errno.EEXIST, "No usable temporary file name found")


def umask_fix_perms( path, umask, unmasked_perms, gid=None ):
    """
    umask-friendly permissions fixing
    """
    perms = unmasked_perms & ~umask
    try:
        st = os.stat( path )
    except OSError as e:
        log.exception( 'Unable to set permissions or group on %s' % path )
        return
    # fix modes
    if stat.S_IMODE( st.st_mode ) != perms:
        try:
            os.chmod( path, perms )
        except Exception as e:
            log.warning( 'Unable to honor umask (%s) for %s, tried to set: %s but mode remains %s, error was: %s' % ( oct( umask ),
                                                                                                                      path,
                                                                                                                      oct( perms ),
                                                                                                                      oct( stat.S_IMODE( st.st_mode ) ),
                                                                                                                      e ) )
    # fix group
    if gid is not None and st.st_gid != gid:
        try:
            os.chown( path, -1, gid )
        except Exception as e:
            try:
                desired_group = grp.getgrgid( gid )
                current_group = grp.getgrgid( st.st_gid )
            except:
                desired_group = gid
                current_group = st.st_gid
            log.warning( 'Unable to honor primary group (%s) for %s, group remains %s, error was: %s' % ( desired_group,
                                                                                                          path,
                                                                                                          current_group,
                                                                                                          e ) )


def docstring_trim(docstring):
    """Trimming python doc strings. Taken from: http://www.python.org/dev/peps/pep-0257/"""
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxint
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxint:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)


def nice_size(size):
    """
    Returns a readably formatted string with the size

    >>> nice_size(100)
    '100 bytes'
    >>> nice_size(10000)
    '9.8 KB'
    >>> nice_size(1000000)
    '976.6 KB'
    >>> nice_size(100000000)
    '95.4 MB'
    """
    words = [ 'bytes', 'KB', 'MB', 'GB', 'TB' ]
    prefix = ''
    try:
        size = float( size )
        if size < 0:
            size = abs( size )
            prefix = '-'
    except:
        return '??? bytes'
    for ind, word in enumerate(words):
        step = 1024 ** (ind + 1)
        if step > size:
            size = size / float(1024 ** ind)
            if word == 'bytes':  # No decimals for bytes
                return "%s%d bytes" % ( prefix, size )
            return "%s%.1f %s" % ( prefix, size, word )
    return '??? bytes'


def size_to_bytes( size ):
    """
    Returns a number of bytes if given a reasonably formatted string with the size
    """
    # Assume input in bytes if we can convert directly to an int
    try:
        return int( size )
    except:
        pass
    # Otherwise it must have non-numeric characters
    size_re = re.compile( '([\d\.]+)\s*([tgmk]b?|b|bytes?)$' )
    size_match = re.match( size_re, size.lower() )
    assert size_match is not None
    size = float( size_match.group(1) )
    multiple = size_match.group(2)
    if multiple.startswith( 't' ):
        return int( size * 1024 ** 4 )
    elif multiple.startswith( 'g' ):
        return int( size * 1024 ** 3 )
    elif multiple.startswith( 'm' ):
        return int( size * 1024 ** 2 )
    elif multiple.startswith( 'k' ):
        return int( size * 1024 )
    elif multiple.startswith( 'b' ):
        return int( size )


def send_mail( frm, to, subject, body, config ):
    """
    Sends an email.
    """
    to = listify( to )
    msg = email_mime_text(  body.encode( 'ascii', 'replace' ) )
    msg[ 'To' ] = ', '.join( to )
    msg[ 'From' ] = frm
    msg[ 'Subject' ] = subject
    if config.smtp_server is None:
        log.error( "Mail is not configured for this Galaxy instance." )
        log.info( msg )
        return
    smtp_ssl = asbool( getattr(config, 'smtp_ssl', False ) )
    if smtp_ssl:
        s = smtplib.SMTP_SSL()
    else:
        s = smtplib.SMTP()
    s.connect( config.smtp_server )
    if not smtp_ssl:
        try:
            s.starttls()
            log.debug( 'Initiated SSL/TLS connection to SMTP server: %s' % config.smtp_server )
        except RuntimeError as e:
            log.warning( 'SSL/TLS support is not available to your Python interpreter: %s' % e )
        except smtplib.SMTPHeloError as e:
            log.error( "The server didn't reply properly to the HELO greeting: %s" % e )
            s.close()
            raise
        except smtplib.SMTPException as e:
            log.warning( 'The server does not support the STARTTLS extension: %s' % e )
    if config.smtp_username and config.smtp_password:
        try:
            s.login( config.smtp_username, config.smtp_password )
        except smtplib.SMTPHeloError as e:
            log.error( "The server didn't reply properly to the HELO greeting: %s" % e )
            s.close()
            raise
        except smtplib.SMTPAuthenticationError as e:
            log.error( "The server didn't accept the username/password combination: %s" % e )
            s.close()
            raise
        except smtplib.SMTPException as e:
            log.error( "No suitable authentication method was found: %s" % e )
            s.close()
            raise
    s.sendmail( frm, to, msg.as_string() )
    s.quit()


def force_symlink( source, link_name ):
    try:
        os.symlink( source, link_name )
    except OSError as e:
        if e.errno == errno.EEXIST:
            os.remove( link_name )
            os.symlink( source, link_name )
        else:
            raise e


def move_merge( source, target ):
    # when using shutil and moving a directory, if the target exists,
    # then the directory is placed inside of it
    # if the target doesn't exist, then the target is made into the directory
    # this makes it so that the target is always the target, and if it exists,
    # the source contents are moved into the target
    if os.path.isdir( source ) and os.path.exists( target ) and os.path.isdir( target ):
        for name in os.listdir( source ):
            move_merge( os.path.join( source, name ), os.path.join( target, name ) )
    else:
        return shutil.move( source, target )


def safe_str_cmp(a, b):
    """safely compare two strings in a timing-attack-resistant manner
    """
    if len(a) != len(b):
        return False
    rv = 0
    for x, y in zip(a, b):
        rv |= ord(x) ^ ord(y)
    return rv == 0

galaxy_root_path = os.path.join(__path__[0], "..", "..", "..")


def galaxy_directory():
    return os.path.abspath(galaxy_root_path)


def parse_int(value, min_val=None, max_val=None, default=None, allow_none=False):
    try:
        value = int(value)
        if min_val is not None and value < min_val:
            return min_val
        if max_val is not None and value > max_val:
            return max_val
        return value
    except ValueError:
        if allow_none:
            if default is None or value == "None":
                return None
        if default:
            return default
        else:
            raise


class ExecutionTimer(object):

    def __init__(self):
        self.begin = time.time()

    def __str__(self):
        elapsed = (time.time() - self.begin) * 1000.0
        return "(%0.3f ms)" % elapsed

if __name__ == '__main__':
    import doctest
    doctest.testmod(sys.modules[__name__], verbose=False)



"""
Dependency management for tools.
"""

import os.path

import logging
log = logging.getLogger( __name__ )

from .resolvers import INDETERMINATE_DEPENDENCY
from .resolvers.galaxy_packages import GalaxyPackageDependencyResolver
from .resolvers.tool_shed_packages import ToolShedPackageDependencyResolver
from .resolvers.conda import CondaDependencyResolver
from galaxy.util import plugin_config

# TODO: Load these from the plugins. Would require a two step initialization of
# DependencyManager - where the plugins are loaded first and then the config
# is parsed and sent through.
EXTRA_CONFIG_KWDS = {
    'conda_prefix': None,
    'conda_exec': None,
    'conda_debug': None,
    'conda_channels': 'r,bioconda',
    'conda_auto_install': False,
    'conda_auto_init': False,
}


def build_dependency_manager( config ):
    if getattr( config, "use_tool_dependencies", False ):
        dependency_manager_kwds = {
            'default_base_path': config.tool_dependency_dir,
            'conf_file': config.dependency_resolvers_config_file,
        }
        for key, default_value in EXTRA_CONFIG_KWDS.iteritems():
            dependency_manager_kwds[key] = getattr(config, key, default_value)
        dependency_manager = DependencyManager( **dependency_manager_kwds )
    else:
        dependency_manager = NullDependencyManager()

    return dependency_manager


class NullDependencyManager( object ):

    def uses_tool_shed_dependencies(self):
        return False

    def dependency_shell_commands( self, requirements, **kwds ):
        return []

    def find_dep( self, name, version=None, type='package', **kwds ):
        return INDETERMINATE_DEPENDENCY


class DependencyManager( object ):
    """
    A DependencyManager attempts to resolve named and versioned dependencies by
    searching for them under a list of directories. Directories should be
    of the form:

        $BASE/name/version/...

    and should each contain a file 'env.sh' which can be sourced to make the
    dependency available in the current shell environment.
    """
    def __init__( self, default_base_path, conf_file=None, **extra_config ):
        """
        Create a new dependency manager looking for packages under the paths listed
        in `base_paths`.  The default base path is app.config.tool_dependency_dir.
        """
        if not os.path.exists( default_base_path ):
            log.warn( "Path '%s' does not exist, ignoring", default_base_path )
        if not os.path.isdir( default_base_path ):
            log.warn( "Path '%s' is not directory, ignoring", default_base_path )
        self.extra_config = extra_config
        self.default_base_path = os.path.abspath( default_base_path )
        self.resolver_classes = self.__resolvers_dict()
        self.dependency_resolvers = self.__build_dependency_resolvers( conf_file )

    def dependency_shell_commands( self, requirements, **kwds ):
        commands = []
        for requirement in requirements:
            log.debug( "Building dependency shell command for dependency '%s'", requirement.name )
            dependency = INDETERMINATE_DEPENDENCY
            if requirement.type in [ 'package', 'set_environment' ]:
                dependency = self.find_dep( name=requirement.name,
                                            version=requirement.version,
                                            type=requirement.type,
                                            **kwds )
            dependency_commands = dependency.shell_commands( requirement )
            if not dependency_commands:
                log.warn( "Failed to resolve dependency on '%s', ignoring", requirement.name )
            else:
                commands.append( dependency_commands )
        return commands

    def uses_tool_shed_dependencies(self):
        return any( map( lambda r: isinstance( r, ToolShedPackageDependencyResolver ), self.dependency_resolvers ) )

    def find_dep( self, name, version=None, type='package', **kwds ):
        log.debug('Find dependency %s version %s' % (name, version))
        for resolver in self.dependency_resolvers:
            dependency = resolver.resolve( name, version, type, **kwds )
            log.debug('Resolver %s returned %s (isnull? %s)' % (resolver.resolver_type, dependency,
                                                                dependency == INDETERMINATE_DEPENDENCY))
            if dependency != INDETERMINATE_DEPENDENCY:
                return dependency
        return INDETERMINATE_DEPENDENCY

    def __build_dependency_resolvers( self, conf_file ):
        if not conf_file:
            return self.__default_dependency_resolvers()
        if not os.path.exists( conf_file ):
            log.debug( "Unable to find config file '%s'", conf_file)
            return self.__default_dependency_resolvers()
        plugin_source = plugin_config.plugin_source_from_path( conf_file )
        return self.__parse_resolver_conf_xml( plugin_source )

    def __default_dependency_resolvers( self ):
        return [
            ToolShedPackageDependencyResolver(self),
            GalaxyPackageDependencyResolver(self),
            GalaxyPackageDependencyResolver(self, versionless=True),
            CondaDependencyResolver(self),
            CondaDependencyResolver(self, versionless=True),
        ]

    def __parse_resolver_conf_xml(self, plugin_source):
        """
        """
        extra_kwds = dict( dependency_manager=self )
        return plugin_config.load_plugins( self.resolver_classes, plugin_source, extra_kwds )

    def __resolvers_dict( self ):
        import galaxy.tools.deps.resolvers
        return plugin_config.plugins_dict( galaxy.tools.deps.resolvers, 'resolver_type' )



import functools
import hashlib
import json
import os.path
import re
import shutil
from sys import platform as _platform
import tempfile

import six
import yaml

from ..deps import commands

# Not sure there are security concerns, lets just fail fast if we are going
# break shell commands we are building.
SHELL_UNSAFE_PATTERN = re.compile(r"[\s\"']")

IS_OS_X = _platform == "darwin"

# BSD 3-clause
CONDA_LICENSE = "http://docs.continuum.io/anaconda/eula"


def conda_link():
    if IS_OS_X:
        url = "https://repo.continuum.io/miniconda/Miniconda-latest-MacOSX-x86_64.sh"
    else:
        url = "https://repo.continuum.io/miniconda/Miniconda-latest-Linux-x86_64.sh"
    return url


def find_conda_prefix(conda_prefix=None):
    """ If supplied conda_prefix is not set, default to the default location
    for Miniconda installs.
    """
    if conda_prefix is None:
        return os.path.join(os.path.expanduser("~"), "miniconda2")
    return conda_prefix


class CondaContext(object):

    def __init__(self, conda_prefix=None, conda_exec=None,
                 shell_exec=None, debug=False, ensure_channels=''):
        conda_exec = conda_exec or commands.which("conda")
        if self.conda_exec:
            conda_exec = os.path.normpath(conda_exec)
        self.conda_exec = conda_exec
        self.debug = debug
        self.shell_exec = shell_exec or commands.shell

        if conda_prefix is None:
            info = self.conda_info()
            if info and "default_prefix" in info:
                conda_prefix = info["default_prefix"]
        if conda_prefix is None:
            conda_prefix = find_conda_prefix(conda_prefix)

        self.conda_prefix = conda_prefix
        if conda_exec is None:
            self.conda_exec = self._bin("conda")
        if ensure_channels:
            if not isinstance(ensure_channels, list):
                ensure_channels = [c for c in ensure_channels.split(",") if c]

            changed = False
            conda_conf = self.load_condarc()
            if "channels" not in conda_conf:
                conda_conf["channels"] = []
            channels = conda_conf["channels"]
            for channel in ensure_channels:
                if channel not in channels:
                    changed = True
                    channels.append(channel)

            if changed:
                self.save_condarc(conda_conf)

    def conda_info(self):
        if self.conda_exec is not None:
            info_out = commands.execute([self.conda_exec, "info", "--json"])
            info = json.loads(info_out)
            return info
        else:
            return None

    def load_condarc(self):
        condarc = self.condarc
        if os.path.exists(condarc):
            with open(condarc, "r") as f:
                return yaml.safe_load(f)
        else:
            return {"channels": ["defaults"]}

    def save_condarc(self, conf):
        condarc = self.condarc
        with open(condarc, "w") as f:
            return yaml.safe_dump(conf, f)

    @property
    def condarc(self):
        return os.path.join(os.path.expanduser("~"), ".condarc")

    def command(self, operation, args):
        if isinstance(args, list):
            args = " ".join(args)
        conda_prefix = self.conda_exec
        if self.debug:
            conda_prefix = "%s --debug"
        return "%s %s %s" % (conda_prefix, operation, args)

    def exec_command(self, operation, args):
        command = self.command(operation, args)
        self.shell_exec(command)

    def exec_create(self, args):
        create_base_args = [
            "-y"
        ]
        create_base_args.extend(args)
        return self.exec_command("create", create_base_args)

    def exec_install(self, args):
        install_base_args = [
            "-y"
        ]
        install_base_args.extend(args)
        return self.exec_command("install", install_base_args)

    def export_list(self, name, path):
        return self.exec_command("list", [
            "--name", name,
            "--export", ">", path
        ])

    def env_path(self, env_name):
        return os.path.join(self.conda_prefix, "envs", env_name)

    def has_env(self, env_name):
        env_path = self.env_path(env_name)
        return os.path.isdir(env_path)

    @property
    def deactivate(self):
        return self._bin("deactivate")

    @property
    def activate(self):
        return self._bin("activate")

    def _bin(self, name):
        return os.path.join(self.conda_prefix, "bin", name)


@six.python_2_unicode_compatible
class CondaTarget(object):

    def __init__(self, package, version=None, channel=None):
        if SHELL_UNSAFE_PATTERN.search(package) is not None:
            raise ValueError("Invalid package [%s] encountered." % package)
        self.package = package
        if version and SHELL_UNSAFE_PATTERN.search(version) is not None:
            raise ValueError("Invalid version [%s] encountered." % version)
        self.version = version
        if channel and SHELL_UNSAFE_PATTERN.search(channel) is not None:
            raise ValueError("Invalid version [%s] encountered." % channel)
        self.channel = channel

    def __str__(self):
        attributes = "package=%s" % self.package
        if self.version is not None:
            attributes = "%s,version=%s" % (self.package, self.version)
        else:
            attributes = "%s,unversioned" % self.package

        if self.channel:
            attributes = "%s,channel=%s" % self.channel

        return "CondaTarget[%s]" % attributes

    @property
    def package_specifier(self):
        """ Return a package specifier as consumed by conda install/create.
        """
        if self.version:
            return "%s=%s" % (self.package, self.version)
        else:
            return self.package

    @property
    def install_environment(self):
        """ The dependency resolution and installation frameworks will
        expect each target to be installed it its own environment with
        a fixed and predictable name given package and version.
        """
        if self.version:
            return "__package__%s@__version__%s" % (self.package, self.version)
        else:
            return "__package__%s@__unversion__" % (self.package)


def hash_conda_packages(conda_packages, conda_target=None):
    """ Produce a unique hash on supplied packages.
    TODO: Ideally we would do this in such a way that preserved environments.
    """
    h = hashlib.new('sha256')
    for conda_package in conda_packages:
        h.update(conda_package.install_environment)
    return h.hexdigest()


def install_conda(conda_context=None):
    conda_context = _ensure_conda_context(conda_context)
    download_cmd = " ".join(commands.download_command(conda_link(), quote_url=True))
    download_cmd = "%s > /tmp/conda.bash" % download_cmd
    install_cmd = "bash /tmp/conda.bash -b -p '%s'" % conda_context.conda_prefix
    full_command = "%s; %s" % (download_cmd, install_cmd)
    return conda_context.shell_exec(full_command)


def install_conda_target(conda_target, conda_context=None):
    """ Install specified target into a its own environment.
    """
    conda_context = _ensure_conda_context(conda_context)
    create_args = [
        "--name", conda_target.install_environment,  # enviornment for package
        conda_target.package_specifier,
    ]
    conda_context.exec_create(create_args)


def is_conda_target_installed(conda_target, conda_context=None):
    conda_context = _ensure_conda_context(conda_context)
    return conda_context.has_env(conda_target.install_environment)


def filter_installed_targets(conda_targets, conda_context=None):
    conda_context = _ensure_conda_context(conda_context)
    installed = functools.partial(is_conda_target_installed,
                                  conda_context=conda_context)
    return filter(installed, conda_targets)


def build_isolated_environment(
    conda_packages,
    path=None,
    copy=False,
    conda_context=None,
):
    """ Build a new environment (or reuse an existing one from hashes)
    for specified conda packages.
    """
    if not isinstance(conda_packages, list):
        conda_packages = [conda_packages]

    # Lots we could do in here, hashing, checking revisions, etc...
    conda_context = _ensure_conda_context(conda_context)
    try:
        hash = hash_conda_packages(conda_packages)
        tempdir = tempfile.mkdtemp(prefix="jobdeps", suffix=hash)
        tempdir_name = os.path.basename(tempdir)

        export_paths = []
        for conda_package in conda_packages:
            name = conda_package.install_environment
            export_path = os.path.join(tempdir, name)
            conda_context.export_list(
                name,
                export_path
            )
            export_paths.append(export_path)
        create_args = ["--unknown", "--offline"]
        if path is None:
            create_args.extend(["--name", tempdir_name])
        else:
            create_args.extend(["--prefix", path])

        if copy:
            create_args.append("--copy")
        for export_path in export_paths:
            create_args.extend([
                "--file", export_path, ">", "/dev/null"
            ])

        if path is not None and os.path.exists(path):
            exit_code = conda_context.exec_install(create_args)
        else:
            exit_code = conda_context.exec_create(create_args)

        return (path or tempdir_name, exit_code)
    finally:
        shutil.rmtree(tempdir)


def requirement_to_conda_targets(requirement, conda_context=None):
    conda_target = None
    if requirement.type == "package":
        conda_target = CondaTarget(requirement.name,
                                   version=requirement.version)
    return conda_target


def requirements_to_conda_targets(requirements, conda_context=None):
    r_to_ct = functools.partial(requirement_to_conda_targets,
                                conda_context=conda_context)
    conda_targets = map(r_to_ct, requirements)
    return [c for c in conda_targets if c is not None]


def _ensure_conda_context(conda_context):
    if conda_context is None:
        conda_context = CondaContext()
    return conda_context


__all__ = [
    'CondaContext',
    'CondaTarget',
    'install_conda',
    'install_conda_target',
    'requirements_to_conda_targets',
]



from galaxy.util import xml_text

DEFAULT_REQUIREMENT_TYPE = "package"
DEFAULT_REQUIREMENT_VERSION = None


class ToolRequirement( object ):
    """
    Represents an external requirement that must be available for the tool to
    run (for example, a program, package, or library).  Requirements can
    optionally assert a specific version.
    """
    def __init__( self, name=None, type=None, version=None ):
        self.name = name
        self.type = type
        self.version = version

    def to_dict( self ):
        return dict(name=self.name, type=self.type, version=self.version)

    @staticmethod
    def from_dict( dict ):
        version = dict.get( "version", None )
        name = dict.get("name", None)
        type = dict.get("type", None)
        return ToolRequirement( name=name, type=type, version=version )


DEFAULT_CONTAINER_TYPE = "docker"


class ContainerDescription( object ):

    def __init__( self, identifier=None, type="docker" ):
        self.identifier = identifier
        self.type = type

    def to_dict( self ):
        return dict(identifier=self.identifier, type=self.type)

    @staticmethod
    def from_dict( dict ):
        identifier = dict["identifier"]
        type = dict.get("type", DEFAULT_CONTAINER_TYPE)
        return ContainerDescription( identifier=identifier, type=type )


def parse_requirements_from_dict( root_dict ):
    requirements = root_dict.get("requirements", [])
    containers = root_dict.get("containers", [])
    return map(ToolRequirement.from_dict, requirements), map(ContainerDescription.from_dict, containers)


def parse_requirements_from_xml( xml_root ):
    """

    >>> from xml.etree import ElementTree
    >>> def load_requirements( contents ):
    ...     contents_document = '''<tool><requirements>%s</requirements></tool>'''
    ...     root = ElementTree.fromstring( contents_document % contents )
    ...     return parse_requirements_from_xml( root )
    >>> reqs, containers = load_requirements('''<requirement>bwa</requirement>''')
    >>> reqs[0].name
    'bwa'
    >>> reqs[0].version is None
    True
    >>> reqs[0].type
    'package'
    >>> reqs, containers = load_requirements('''<requirement type="binary" version="1.3.3">cufflinks</requirement>''')
    >>> reqs[0].name
    'cufflinks'
    >>> reqs[0].version
    '1.3.3'
    >>> reqs[0].type
    'binary'
    """
    requirements_elem = xml_root.find( "requirements" )

    requirement_elems = []
    if requirements_elem is not None:
        requirement_elems = requirements_elem.findall( 'requirement' )

    requirements = []
    for requirement_elem in requirement_elems:
        name = xml_text( requirement_elem )
        type = requirement_elem.get( "type", DEFAULT_REQUIREMENT_TYPE )
        version = requirement_elem.get( "version", DEFAULT_REQUIREMENT_VERSION )
        requirement = ToolRequirement( name=name, type=type, version=version )
        requirements.append( requirement )

    container_elems = []
    if requirements_elem is not None:
        container_elems = requirements_elem.findall( 'container' )

    containers = []
    for container_elem in container_elems:
        identifier = xml_text( container_elem )
        type = container_elem.get( "type", DEFAULT_CONTAINER_TYPE )
        container = ContainerDescription( identifier=identifier, type=type )
        containers.append( container )

    return requirements, containers



""" API for this module containing functionality related to the tool box.
"""

from .panel import panel_item_types
from .panel import ToolSection
from .panel import ToolSectionLabel

from .base import AbstractToolBox

__all__ = [
    "ToolSection",
    "ToolSectionLabel",
    "panel_item_types",
    "AbstractToolBox",
]



from abc import ABCMeta
from abc import abstractmethod

import os.path

from ...metrics import formatting


INSTRUMENT_FILE_PREFIX = "__instrument"


class InstrumentPlugin( object ):
    """ A plugin describing how to instrument Galaxy jobs and retrieve metrics
    from this instrumentation.
    """
    __metaclass__ = ABCMeta
    formatter = formatting.JobMetricFormatter()

    @property
    @abstractmethod
    def plugin_type( self ):
        """ Short string providing labelling this plugin """

    def pre_execute_instrument( self, job_directory ):
        """ Optionally return one or more commands to instrument job. These
        commands will be executed on the compute server prior to the job
        running.
        """
        return None

    def post_execute_instrument( self, job_directory ):
        """ Optionally return one or more commands to instrument job. These
        commands will be executed on the compute server after the tool defined
        command is ran.
        """
        return None

    @abstractmethod
    def job_properties( self, job_id, job_directory ):
        """ Collect properties for this plugin from specified job directory.
        This method will run on the Galaxy server and can assume files created
        in job_directory with pre_execute_instrument and
        post_execute_instrument are available.
        """

    def _instrument_file_name( self, name ):
        """ Provide a common pattern for naming files used by instrumentation
        plugins - to ease their staging out of remote job directories.
        """
        return "%s_%s_%s" % ( INSTRUMENT_FILE_PREFIX, self.plugin_type, name )

    def _instrument_file_path( self, job_directory, name ):
        return os.path.join( job_directory, self._instrument_file_name( name ) )



import os
import shutil

from ..instrumenters import InstrumentPlugin
from ...metrics import formatting
from ..collectl import subsystems
from ..collectl import cli
from ..collectl import processes

from galaxy import util
from galaxy.util import directory_hash

import logging
log = logging.getLogger( __name__ )

# By default, only grab statistics for user processes (as identified by
# username).
DEFAULT_PROCFILT_ON = "username"
DEFAULT_SUBSYSTEMS = "process"
# Set to zero to flush every collection.
DEFAULT_FLUSH_INTERVAL = "0"

FORMATTED_RESOURCE_TITLES = {
    "PCT": "Percent CPU Usage",
    "RSYS": "Disk Reads",
    "WSYS": "Disk Writes",
}

EMPTY_COLLECTL_FILE_MESSAGE = "Skipping process summary due to empty file... job probably did not run long enough for collectl to gather data."


class CollectlFormatter( formatting.JobMetricFormatter ):

    def format( self, key, value ):
        if key == "pid":
            return ( "Process ID", int( value ) )
        elif key == "raw_log_path":
            return ( "Relative Path of Full Collectl Log", value )
        elif key == "process_max_AccumT":
            return ( "Job Runtime (System+User)", formatting.seconds_to_str( float( value ) ) )
        else:
            _, stat_type, resource_type = key.split( "_", 2 )
            if resource_type.startswith( "Vm"):
                value_str = "%s KB" % int( value )
            elif resource_type in [ "RSYS", "WSYS" ] and stat_type in [ "count", "max", "sum" ]:
                value_str = "%d (# system calls)" % int( value )
            else:
                value_str = str( value )
            resource_title = FORMATTED_RESOURCE_TITLES.get( resource_type, resource_type )
            return ( "%s (%s)" % ( resource_title, stat_type ), value_str )


class CollectlPlugin( InstrumentPlugin ):
    """ Run collectl along with job to capture system and/or process data
    according to specified collectl subsystems.
    """
    plugin_type = "collectl"
    formatter = CollectlFormatter()

    def __init__( self, **kwargs ):
        self.__configure_paths( kwargs )
        self.__configure_subsystems( kwargs )
        saved_logs_path = kwargs.get( "saved_logs_path", "" )
        if "app" in kwargs:
            log.debug("Found path for saved logs: %s" % saved_logs_path)
            saved_logs_path = kwargs[ "app" ].config.resolve_path( saved_logs_path )
        self.saved_logs_path = saved_logs_path
        self.__configure_collectl_recorder_args( kwargs )
        self.summarize_process_data = util.asbool( kwargs.get( "summarize_process_data", True ) )
        self.log_collectl_program_output = util.asbool( kwargs.get( "log_collectl_program_output", False ) )
        if self.summarize_process_data:
            if subsystems.get_subsystem( "process" ) not in self.subsystems:
                raise Exception( "Collectl plugin misconfigured - cannot summarize_process_data without process subsystem being enabled." )

            process_statistics = kwargs.get( "process_statistics", None )
            # None will let processes module use default set of statistics
            # defined there.
            self.process_statistics = processes.parse_process_statistics( process_statistics )

    def pre_execute_instrument( self, job_directory ):
        commands = []
        # Capture PID of process so we can walk its ancestors when building
        # statistics for the whole job.
        commands.append( '''echo "$$" > '%s' ''' % self.__pid_file( job_directory ) )
        # Run collectl in record mode to capture process and system level
        # statistics according to supplied subsystems.
        commands.append( self.__collectl_record_command( job_directory ) )
        return commands

    def post_execute_instrument( self, job_directory ):
        commands = []
        # collectl dies when job script completes, perhaps capture pid of
        # collectl above and check if it is still alive to allow tracking if
        # collectl ran successfully through the whole job.
        return commands

    def job_properties( self, job_id, job_directory ):
        pid = open( self.__pid_file( job_directory ), "r" ).read().strip()
        contents = os.listdir( job_directory )
        try:
            rel_path = filter( self._is_instrumented_collectl_log, contents )[ 0 ]
            path = os.path.join( job_directory, rel_path )
        except IndexError:
            message = "Failed to find collectl log in directory %s, files were %s" % ( job_directory, contents )
            raise Exception( message )

        properties = dict(
            pid=int( pid ),
        )

        if self.saved_logs_path:
            destination_rel_dir = os.path.join( *directory_hash.directory_hash_id( job_id ) )
            destination_rel_path = os.path.join( destination_rel_dir, rel_path )
            destination_path = os.path.join( self.saved_logs_path, destination_rel_path )
            destination_dir = os.path.dirname( destination_path )
            if not os.path.isdir( destination_dir ):
                os.makedirs( destination_dir )
            shutil.copyfile( path, destination_path )
            properties[ "raw_log_path" ] = destination_rel_path

        if self.summarize_process_data:
            # Run collectl in playback and generate statistics of interest
            summary_statistics = self.__summarize_process_data( pid, path )
            for statistic, value in summary_statistics:
                properties[ "process_%s" % "_".join( statistic ) ] = value

        return properties

    def __configure_paths( self, kwargs ):
        # 95% of time I would expect collectl to just be installed with apt or
        # yum, but if it is manually installed on not on path, allow
        # configuration of explicit path - and allow path to be different
        # between galaxy job handler (local_collectl_path) and compute node
        # (remote_collectl_path).
        collectl_path = kwargs.get( "collectl_path", "collectl" )
        self.remote_collectl_path = kwargs.get( "remote_collectl_path", collectl_path )
        self.local_collectl_path = kwargs.get( "local_collectl_path", collectl_path )

    def __configure_subsystems( self, kwargs ):
        raw_subsystems_str = kwargs.get( "subsystems", DEFAULT_SUBSYSTEMS )
        raw_subsystems = util.listify( raw_subsystems_str, do_strip=True )
        self.subsystems = map( subsystems.get_subsystem, raw_subsystems )

    def __configure_collectl_recorder_args( self, kwargs ):
        collectl_recorder_args = kwargs.copy()

        # Allow deployer to configure separate system and process intervals,
        # but if they specify just one - use it for both. Thinking here is this
        # plugin's most useful feature is the process level information so
        # this is likely what the deployer is attempting to configure.
        if "interval" in kwargs and "interval2" not in kwargs:
            collectl_recorder_args[ "interval2" ] = kwargs[ "interval"]

        if "flush" not in kwargs:
            collectl_recorder_args[ "flush" ] = DEFAULT_FLUSH_INTERVAL

        procfilt_on = kwargs.get( "procfilt_on", DEFAULT_PROCFILT_ON ).lower()
        # Calculate explicit arguments, rest can just be passed through from
        # constructor arguments.
        explicit_args = dict(
            collectl_path=self.remote_collectl_path,
            procfilt=procfilt_argument( procfilt_on ),
            subsystems=self.subsystems,
        )
        collectl_recorder_args.update( explicit_args )
        self.collectl_recorder_args = collectl_recorder_args

    def __summarize_process_data( self, pid, collectl_log_path ):
        playback_cli_args = dict(
            collectl_path=self.local_collectl_path,
            playback_path=collectl_log_path,
            sep="9"
        )
        if not os.stat( collectl_log_path ).st_size:
            log.debug( EMPTY_COLLECTL_FILE_MESSAGE )
            return [ ]

        playback_cli = cli.CollectlCli( **playback_cli_args )
        return processes.generate_process_statistics( playback_cli, pid, self.process_statistics )

    def __collectl_recorder_cli( self, job_directory ):
        cli_args = self.collectl_recorder_args.copy()
        cli_args[ "destination_path" ] = self._instrument_file_path( job_directory, "log" )
        return cli.CollectlCli( **cli_args )

    def __collectl_record_command( self, job_directory ):
        collectl_cli = self.__collectl_recorder_cli( job_directory )
        if self.log_collectl_program_output:
            redirect_to = self._instrument_file_path( job_directory, "program_output" )
        else:
            redirect_to = "/dev/null"
        return "%s > %s 2>&1 &" % (
            collectl_cli.build_command_line(),
            redirect_to,
        )

    def __pid_file( self, job_directory ):
        return self._instrument_file_path( job_directory, "pid" )

    def _is_instrumented_collectl_log( self, filename ):
        prefix = self._instrument_file_name( "log" )
        return filename.startswith( prefix ) and filename.endswith( ".raw.gz" )


def procfilt_argument( procfilt_on ):
    if procfilt_on == "username":
        return "U$USER"
    elif procfilt_on == "uid":
        return "u$UID"
    else:
        # Ensure it is empty of None
        if procfilt_on or procfilt_on.lower() != "none":
            raise Exception( "Invalid procfilt_on argument encountered")
        return ""


__all__ = [ 'CollectlPlugin' ]



""" Module used to blend ini, environment, and explicit dictionary properties
to determine application configuration. Some hard coded defaults for Galaxy but
this should be reusable by tool shed and pulsar as well.
"""
import os
import os.path


def load_app_properties(
    kwds={},
    ini_file=None,
    ini_section="app:main",
    config_prefix="GALAXY_CONFIG_"
):
    properties = kwds.copy() if kwds else {}
    if ini_file:
        defaults = {
            'here': os.path.dirname(os.path.abspath(ini_file)),
            '__file__': os.path.abspath(ini_file)
        }
        # Mimic the way loadwsgi loads configuration files - needed for
        # correctness given the very specific ways interpolation is handled.
        from galaxy.util.pastescript.loadwsgi import NicerConfigParser
        parser = NicerConfigParser(ini_file, defaults=defaults)
        parser.optionxform = str  # Don't lower-case keys
        with open(ini_file) as f:
            parser.read_file(f)

        properties.update( dict( parser.items( ini_section ) ) )

    override_prefix = "%sOVERRIDE_" % config_prefix
    for key in os.environ:
        if key.startswith( override_prefix ):
            config_key = key[ len( override_prefix ): ].lower()
            properties[ config_key ] = os.environ[ key ]
        elif key.startswith( config_prefix ):
            config_key = key[ len( config_prefix ): ].lower()
            if config_key not in properties:
                properties[ config_key ] = os.environ[ key ]

    return properties



import re

from ..instrumenters import InstrumentPlugin
from ...metrics import formatting

import logging
log = logging.getLogger( __name__ )


class EnvFormatter( formatting.JobMetricFormatter ):

    def format( self, key, value ):
        return ( "%s (runtime environment variable)" % key, value )


class EnvPlugin( InstrumentPlugin ):
    """ Instrumentation plugin capable of recording all or specific environment
    variables for a job at runtime.
    """
    plugin_type = "env"
    formatter = EnvFormatter()

    def __init__( self, **kwargs ):
        variables_str = kwargs.get( "variables", None )
        if variables_str:
            variables = [ v.strip() for v in variables_str.split(",") ]
        else:
            variables = None
        self.variables = variables

    def pre_execute_instrument( self, job_directory ):
        """ Use env to dump all environment variables to a file.
        """
        return "env > '%s'" % self.__env_file( job_directory )

    def post_execute_instrument( self, job_directory ):
        return None

    def job_properties( self, job_id, job_directory ):
        """ Recover environment variables dumped out on compute server and filter
        out specific variables if needed.
        """
        variables = self.variables

        properties = {}
        env_string = ''.join( open( self.__env_file( job_directory ) ).readlines() )
        while env_string:
            # Check if the next lines contain a shell function.
            # We use '\n\}\n' as regex termination because shell
            # functions can be nested.
            # We use the non-greedy '.+?' because of re.DOTALL .
            m = re.match( '([^=]+)=(\(\) \{.+?\n\})\n', env_string, re.DOTALL )
            if m is None:
                m = re.match( '([^=]+)=(.*)\n', env_string )
            if m is None:
                # Some problem recording or reading back env output.
                message_template = "Problem parsing env metric output for job %s - properties will be incomplete"
                message = message_template % job_id
                log.debug( message )
                break
            (var, value) = m.groups()
            if not variables or var in variables:
                properties[ var ] = value
            env_string = env_string[m.end():]

        return properties

    def __env_file( self, job_directory ):
        return self._instrument_file_path( job_directory, "vars" )

__all__ = [ 'EnvPlugin' ]
